(************************************************************************
 Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de
	http://home.t-online.de/home/Robert.Rossmair/

 Module:        RRScales

 Version:	0.91, 10-Jul-98

 Description:

 History:
    16-Feb-95: version 0.1
    03-Jun-98: version 0.2
    24-Jun-98: version 0.9
        TScale.ShowAxis True by default
        TScale.Changed virtual
      New: TControlScale, TrrScaleControl
    10-Jul-98: version 0.91
        DrawGrid

 Compatibility:	Delphi 2 & 3

 Copyright  1995-1999  Robert Rossmair

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License (file COPYING.Lib) along with this library; if not, write
to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
MA 02139, USA.
 ************************************************************************)
unit RRScales;

interface

uses
  Windows,
{$IFNDEF NoMath}
  Math,
{$ENDIF}
  RRMath, Graphics, SysUtils, Classes, Controls;

type
  TScaleTextPos = (tpBottomRight, tpTopLeft);
  TMinScaleLineDistance = 8..High(Integer);
  TDrawGridEvent = procedure(Sender: TObject; ACanvas: TCanvas; Points: array of TPoint) of object;

  TScale = class(TPersistent)
  private
    FStartValue: Extended;
    FStartPoint: TPoint;
    FEndValue: Extended;
    FEndPoint: TPoint;
    FFirstGradValue: Extended;
    FExcept: Integer;	{ scale line to be excluded from drawing; -1: none; 0: first and so forth }
    FValueStep: Extended;
    FFirstGradX: TFixed;
    FFirstGradY: TFixed;
    FStepX: TFixed;
    FStepY: TFixed;
    FFormatStr: string;
    FFont: TFont;
    FOnChange: TNotifyEvent;
    FOnDrawGrid: TDrawGridEvent;
    FUpdateCount: Integer;
    FMinScaleLineDistance: TMinScaleLineDistance;
    FMinTextDistance: Integer; // in space char. width units
    FHorzTextOffset: Integer;
    FTextHeight: Integer;
    FTextPos: TScaleTextPos;
    FVisible: Boolean;
    FShowAxis: Boolean;
    FVertical: boolean;
    procedure Calc;
    procedure ComputeValueStep;
    function GetFirstGradValue: Extended;
    function GetLastGradValue: Extended;
    function GetHeight: Integer;
    function GetTextExtent(S: string): TSize;
    function GetWidth: Integer;
    function LastGraduation(FirstGradValue: Extended): Integer;
    procedure SetFirstGradValue(Value: Extended);
    procedure SetFont(Value: TFont);
    procedure SetStartPoint(Value: TPoint);
    procedure SetStartPointX(Value: Longint);
    procedure SetStartPointY(Value: Longint);
    procedure SetEndPoint(Value: TPoint);
    procedure SetEndPointX(Value: Longint);
    procedure SetEndPointY(Value: Longint);
    procedure SetStartValue(Value: Extended);
    procedure SetEndValue(Value: Extended);
    procedure SetMinScaleLineDistance(Value: TMinScaleLineDistance);
  protected
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Changed; virtual;
    procedure DoDrawGrid(ACanvas: TCanvas; Points: array of TPoint); dynamic;
    procedure FontChange(Sender: TObject);
    procedure SetVisible(Value: Boolean);
    property Font: TFont read FFont write SetFont;
    property MinScaleLineDistance: TMinScaleLineDistance read FMinScaleLineDistance write SetMinScaleLineDistance default 24;
    property StartPointX: Longint read FStartPoint.X write SetStartPointX;
    property StartPointY: Longint read FStartPoint.Y write SetStartPointY;
    property EndPointX: Longint read FEndPoint.X write SetEndPointX;
    property EndPointY: Longint read FEndPoint.Y write SetEndPointY;
    procedure SetShowAxis(Value: Boolean);
    property Visible: Boolean read FVisible write SetVisible default True;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnDrawGrid: TDrawGridEvent read FOnDrawGrid write FOnDrawGrid;
  public
    constructor Create;
    procedure Draw(ACanvas: TCanvas);
    function GetMaxTextExtent: TSize;
    procedure SetSection(NewStartValue, NewEndValue: Extended);
    property Height: Integer read GetHeight;
    property Width: Integer read GetWidth;
    property StartPoint: TPoint read FStartPoint write SetStartPoint;
    property EndPoint: TPoint read FEndPoint write SetEndPoint;
    property StartValue: Extended read FStartValue write SetStartValue;
    property EndValue: Extended read FEndValue write SetEndValue;
    property ShowAxis: Boolean read FShowAxis write SetShowAxis default True;
    property TextPosition: TScaleTextPos read FTextPos write FTextPos;
  end;

  TControlScale = class(TScale)
  private
    FControl: TControl;
  protected
    constructor Create(AControl: TControl);
    procedure Changed; override;
  published
    property Font;
    property MinScaleLineDistance;
    property StartPointX default 8;
    property StartPointY;
    property EndPointX default 208;
    property EndPointY;
    property StartValue;
    property EndValue;
    property Visible;
    property OnChange;
  end;

  TrrScaleControl = class(TGraphicControl)
  private
    FScale: TScale;
    procedure SetScale(Value: TScale);
  protected
    destructor Destroy; override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Scale: TScale read FScale write SetScale;
  end;

procedure Register;

implementation

const
  SZeroLengthSection = 'Cannot scale range of zero length';

constructor TScale.Create;
begin
  inherited Create;
  FStartValue := 0;
  FEndValue := 100;
  FEndPoint := Point(100, 0);
  FMinScaleLineDistance := 24;
  FMinTextDistance := 4;
  FValueStep := 10;
  FExcept := -1;
  FFont := TFont.Create;
  FFont.Style := [];
  FFont.Size := 8;
  FFont.Name := 'Times New Roman';
  FFont.OnChange := FontChange;
  FShowAxis := True;
  FVisible := True;
  FontChange(Self);
end;

destructor TScale.Destroy;
begin
  FFont.Free;
  inherited;
end;

procedure TScale.Assign(Source: TPersistent);
var
  Src: TScale absolute Source;
begin
  if Source is TScale then
  begin
    BeginUpdate;
    try
      Font                 := Src.Font;
      MinScaleLineDistance := Src.MinScaleLineDistance;
      ShowAxis             := Src.ShowAxis;
      Visible              := Src.Visible;
    finally
      EndUpdate;
    end;
  end
  else inherited Assign(Source);
end;

procedure TScale.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TScale.EndUpdate;
begin
  Dec(FUpdateCount);
  if FUpdateCount = 0 then Changed;
end;

procedure TScale.Calc;
var
  Step: Extended;
  dX: integer;
  dY: integer;
  L: word;
begin
  dX := FEndPoint.X-FStartPoint.X;
  dY := FEndPoint.Y-FStartPoint.Y;
  if (dX = 0) and (dY = 0) then Exit;
  FVertical := Abs(dX) <= Abs(dY);
  ComputeValueStep;
  L := uMax(Abs(dX), Abs(dY));
  Step := FValueStep * L * 65536.0 / (FEndValue-FStartValue);
  if FVertical then
  begin
    Longint(FStepY) := Round(Step * Sign(dY));
    Longint(FStepX) := Round(Longint(FStepY)/dY*dX);
  end else
  begin
    Longint(FStepX) := Round(Step * Sign(dX));
    Longint(FStepY) := Round(Longint(FStepX)/dX*dY);
  end;
  SetFirstGradValue(GetFirstGradValue);
end;

procedure TScale.Changed;
begin
  if FVisible then Calc;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TScale.ComputeValueStep;
type
  TScaleStep = (ss1, ss2, ss25, ss5);
const
  lg25 = 0.397940008672038;	{ log10(2.5) }
  lg5 = 0.698970004336019;	{ log10(5) }
  LogFraction: array[TScaleStep] of Extended = (0, lg_2, lg25, lg5);
var
  D, Dmax, Precision: Integer;
  i, L: Integer;
  LogInt: Integer;
  Log, LogFrac: Extended;
  ValueStep: Extended;
  Temp: string;
  Size: TSize;
  ScaleStep: TScaleStep;
label
  Start;
begin
  if (FStartPoint.X = FEndPoint.X) and (FStartPoint.Y = FEndPoint.Y) then Exit;
  Dmax := uMax(Abs(FEndPoint.X-FStartPoint.X), Abs(FEndPoint.Y-FStartPoint.Y));
  ValueStep := (FEndValue-FStartValue) * FMinScaleLineDistance/Dmax;
  Log := log2(Abs(ValueStep))/ld_10;
  LogInt := Floor(Log);
  LogFrac := Log - LogInt;
  ScaleStep := ss1;
  if LogFrac > lg5 then Inc(LogInt) else
  if LogFrac > lg25 then ScaleStep := ss5 else
  if LogFrac > lg_2 then ScaleStep := ss25 else
  if LogFrac > 0 then ScaleStep := ss2;
Start:
  Log := LogInt + LogFraction[ScaleStep];
  if ValueStep < 0
    then FValueStep :=-exp2(Log*ld_10)
    else FValueStep := exp2(Log*ld_10);
  Precision := Ceil(-Log);
  if ScaleStep = ss25 then Inc(Precision);
  if Precision < 0 then Precision := 0;
  FFormatStr := '%.'+IntToStr(Precision)+'f';
  if FVertical then Size := GetMaxTextExtent else
  begin
    Temp := FFormatStr;
    L := Length(FFormatStr);
    SetLength(FFormatStr, L+FMinTextDistance);
    for i := L+1 to L+FMinTextDistance do FFormatStr[i] := ' ';
    Size := GetMaxTextExtent;
    FFormatStr := Temp;
  end;
  if FVertical
    then D := Size.cY
    else D := Size.cX;
  if D > FValueStep/(FEndValue-FStartValue)*DMax then
  begin
    if ScaleStep = ss5 then
    begin
      ScaleStep := ss1;
      Inc(LogInt);
    end else Inc(ScaleStep);
    goto Start;
  end;
end;

function TScale.GetFirstGradValue: Extended;
begin
  Result := Trunc(FStartValue/FValueStep) * FValueStep;
  if Abs(FEndValue-Result) > Abs(FEndValue-FStartValue) then
    Result := Result + FValueStep;
end;

function TScale.GetLastGradValue: Extended;
begin
  Result := GetFirstGradValue;
  Result := Result + LastGraduation(Result) * FValueStep;
end;

function TScale.GetTextExtent(S: string): TSize;
var
  DC: hDC;
  F: hFont;
begin
  DC := GetDC(0);
  try
    F := SelectObject(DC, Font.Handle);
    try
      Result.cX := 0;
      Result.cY := 0;
      Windows.GetTextExtentPoint32(DC, PChar(S), Length(S), Result);
    finally
      SelectObject(DC, F);
    end;
  finally
    ReleaseDC(0, DC);
  end;
end;

function TScale.GetMaxTextExtent: TSize;
var
  DC: hDC;
  F: hFont;
  S1, S2: string;
  S: TSize;
begin
  S1 := Format(FFormatStr, [GetFirstGradValue]);
  S2 := Format(FFormatStr, [GetLastGradValue]);
  DC := GetDC(0);
  try
    F := SelectObject(DC, Font.Handle);
    try
      S.cX := 0;
      S.cY := 0;
      Result.cX := 0;
      Result.cY := 0;
      Windows.GetTextExtentPoint32(DC, PChar(S1), Length(S1), S);
      Windows.GetTextExtentPoint32(DC, PChar(S2), Length(S2), Result);
      if S.cX > Result.cX then Result := S;
    finally
      SelectObject(DC, F);
    end;
  finally
    ReleaseDC(0, DC);
  end;
end;

procedure TScale.SetFirstGradValue(Value: Extended);
var
  C: Extended;
begin
  FFirstGradValue := Value;
  C := 65536.0 * (FFirstGradValue-FStartValue) / (FEndValue-FStartValue);
  Longint(FFirstGradX) := Round((FEndPoint.X-FStartPoint.X) * C);
  Longint(FFirstGradY) := Round((FEndPoint.Y-FStartPoint.Y) * C);
end;

function TScale.LastGraduation(FirstGradValue: Extended): Integer;
begin
  Result := Trunc((FEndValue-FirstGradValue)/FValueStep);
end;

procedure TScale.DoDrawGrid(ACanvas: TCanvas; Points: array of TPoint);
begin
  if Assigned(FOnDrawGrid) then FOnDrawGrid(Self, ACanvas, Points);
end;

procedure TScale.Draw(ACanvas: TCanvas);
type
  TPointArray = array[0..9999999] of TPoint;
var
  P: ^TPointArray;
  i, k, C, Th, Tw, Ty: Integer;
  X, Y: Longint;
  U: TFixed absolute X;
  V: TFixed absolute Y;
  Value: Extended;
  S: string;
  OldFont: TFont;
  OldPen: TPen;
begin
  if not FVisible then Exit;
  SetBkMode(ACanvas.Handle, TRANSPARENT);
  OldFont := TFont.Create;
  try
    OldFont.Assign(ACanvas.Font);
    try
      ACanvas.Font := Font;
      OldPen := TPen.Create;
      try
        OldPen.Assign(ACanvas.Pen);
        try
          ACanvas.Pen.Color := Font.Color;
          ACanvas.Pen.Mode  := pmCopy;
          ACanvas.Pen.Style := psSolid;
          if FShowAxis then
          begin
            ACanvas.MoveTo(FStartPoint.X, FStartPoint.Y);
            ACanvas.LineTo(FEndPoint.X, FEndPoint.Y);
            ACanvas.Pixels[FEndPoint.X, FEndPoint.Y] := Font.Color;
          end;
          Value := FFirstGradValue;
          U := FFirstGradX;	Inc(U.Value, FStartPoint.X);
          V := FFirstGradY;	Inc(V.Value, FStartPoint.Y);
          k := LastGraduation(FFirstGradValue);
          GetMem(P, (k+1) * SizeOf(TPoint));
          try
            for i := 0 to k do
            begin
              P[i].X := U.Value;
              P[i].Y := V.Value;
              if i <> FExcept then
              begin
                S := Format(FFormatStr, [Value]);
                Th := ACanvas.TextHeight(S);
                Ty := (Th+1) div 2;
                Tw := ACanvas.TextWidth(S);
                C := Th div 3;
                if FTextPos = tpTopLeft then C := -C;
                with ACanvas do
                if FVertical then
                begin
                  // draw scale line
                  MoveTo(U.Value+C, V.Value);
                  LineTo(U.Value, V.Value);
                  // write scale line label
                  if FTextPos = tpTopLeft
                    then TextOut(U.Value-Ty-Tw, V.Value-Ty, S)
                    else TextOut(U.Value+Ty, V.Value-Ty, S);
                end else
                begin
                  // draw scale line
                  MoveTo(U.Value, V.Value+C);
                  LineTo(U.Value, V.Value);
                  // write scale line label
                  if FTextPos = tpTopLeft
                    then TextOut(U.Value+FHorzTextOffset-Tw, V.Value-Ty-Th, S)
                    else TextOut(U.Value-FHorzTextOffset, V.Value+Ty, S);
                end;
              end;
              Value := Value + FValueStep;
              Inc(X, Longint(FStepX));
              Inc(Y, Longint(FStepY));
            end;
            DoDrawGrid(ACanvas, Slice(P^, k+1));
          finally
            FreeMem(P);
          end;
        finally
          ACanvas.Pen := OldPen;
        end;
      finally
        OldPen.Free;
      end;
    finally
      ACanvas.Font := OldFont;
    end;
  finally
    OldFont.Free;
  end;
end;

function TScale.GetHeight: Integer;
begin
  Result := Abs(FEndPoint.Y-FStartPoint.Y)+1;
  if not FVertical then Inc(Result, FTextHeight*3 div 2);
end;

function TScale.GetWidth: Integer;
var
  Emax: TSize;
begin
  Result := Abs(FEndPoint.X-FStartPoint.X)+1;
  if FVertical then
  begin
    Emax := GetMaxTextExtent;
    Inc(Result, Emax.cX + (Emax.cY+1) div 2);
  end;
end;

procedure TScale.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TScale.SetMinScaleLineDistance(Value: TMinScaleLineDistance);
begin
  if FMinScaleLineDistance <> Value then
  begin
    BeginUpdate;
    FMinScaleLineDistance := Value;
    EndUpdate;
  end;
end;

procedure TScale.SetStartPoint(Value: TPoint);
begin
  if (FStartPoint.X <> Value.X)
  or (FStartPoint.Y <> Value.Y) then
  begin
    BeginUpdate;
    FStartPoint := Value;
    EndUpdate;
  end;
end;

procedure TScale.SetStartPointX(Value: Longint);
begin
  SetStartPoint(Point(Value, FStartPoint.Y));
end;

procedure TScale.SetStartPointY(Value: Longint);
begin
  SetStartPoint(Point(FStartPoint.X, Value));
end;

procedure TScale.SetEndPoint(Value: TPoint);
begin
  if (FEndPoint.X <> Value.X)
  or (FEndPoint.Y <> Value.Y) then
  begin
    BeginUpdate;
    FEndPoint := Value;
    EndUpdate;
  end;
end;

procedure TScale.SetEndPointX(Value: Longint);
begin
  SetEndPoint(Point(Value, FEndPoint.Y));
end;

procedure TScale.SetEndPointY(Value: Longint);
begin
  SetEndPoint(Point(FEndPoint.X, Value));
end;

procedure TScale.SetStartValue(Value: Extended);
begin
  SetSection(Value, EndValue);
end;

procedure TScale.SetEndValue(Value: Extended);
begin
  SetSection(StartValue, Value);
end;

procedure TScale.SetSection(NewStartValue, NewEndValue: Extended);
begin
  if NewStartValue = NewEndValue then raise Exception.Create(SZeroLengthSection);
  BeginUpdate;
  FStartValue := NewStartValue;
  FEndValue := NewEndValue;
  EndUpdate;
end;

procedure TScale.SetShowAxis(Value: Boolean);
begin
  if FShowAxis <> Value then
  begin
    BeginUpdate;
    FShowAxis := Value;
    EndUpdate;
  end;
end;

procedure TScale.SetVisible(Value: Boolean);
begin
  if FVisible <> Value then
  begin
    BeginUpdate;
    FVisible := Value;
    EndUpdate;
  end;
end;

procedure TScale.FontChange(Sender: TObject);
var
  Size: TSize;
begin
  Size := GetTextExtent('3');
  FHorzTextOffset := Size.cX div 2;
  FTextHeight := Size.cY;
  Changed;
end;

constructor TControlScale.Create(AControl: TControl);
begin
  FControl := AControl;
  inherited Create;
  StartPointX := 8;
  EndPointX := 208;
end;

procedure TControlScale.Changed;
begin
  inherited Changed;
  FControl.Invalidate;
end;

constructor TRRScaleControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 224, 20);
  FScale := TControlScale.Create(Self);
end;

destructor TRRScaleControl.Destroy;
begin
  FScale.Free;
  inherited Destroy;
end;

procedure TRRScaleControl.Paint;
begin
  Canvas.Brush.Color := Color;
  FScale.Draw(Canvas);
end;

procedure TRRScaleControl.SetScale(Value: TScale);
begin
  FScale.Assign(Value);
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}
procedure Register;
begin
  RegisterComponents(PalettePage, [TrrScaleControl]);
end;

end.
